-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thStateControlSystem
local thFillTypeManager = thMain.thFillTypeManager
local thInputManager = thMain.thInputManager
local thI18n = thMain.i18n
THStorageTransfer = {}
local THStorageTransfer_mt = THUtils.createClass(THStorageTransfer, THAnimatedProcessor)
InitObjectClass(THStorageTransfer, "THStorageTransfer")
local debugFlagId = THDebugUtil.createDebugFlag("THStorageTransfer")
THStorageTransfer.debugFlagId = debugFlagId
THStorageTransfer.STATUS = THUtils.copyTable(THAnimatedProcessor.STATUS)
THStorageTransfer.STATUS.INPUT_EMPTY = THAnimatedProcessor.MAX_STATUS + 1
THStorageTransfer.STATUS.OUTPUT_FULL = THAnimatedProcessor.MAX_STATUS + 2
THStorageTransfer.STATUS_XML_KEY = THUtils.copyTable(THAnimatedProcessor.STATUS_XML_KEY)
THStorageTransfer.STATUS_XML_KEY[THStorageTransfer.STATUS.INPUT_EMPTY] = "inputEmpty"
THStorageTransfer.STATUS_XML_KEY[THStorageTransfer.STATUS.OUTPUT_FULL] = "outputFull"
THStorageTransfer.STATUS_XML_INDEX = {}
for statusIndex, statusXMLKey in pairs(THStorageTransfer.STATUS_XML_KEY) do
    THStorageTransfer.STATUS_XML_INDEX[statusXMLKey:upper()] = statusIndex
end
THStorageTransfer.STATUS_ID = THUtils.copyTable(THAnimatedProcessor.STATUS_ID)
THStorageTransfer.MAX_STATUS = THStorageTransfer.STATUS.OUTPUT_FULL
for statusId, statusIndex in pairs(THStorageTransfer.STATUS) do
    THStorageTransfer.STATUS_ID[statusIndex] = statusId
end
local getCustomData = function(self)
    return THAnimatedProcessor.getCustomData(self)
end
THStorageTransfer.getCustomData = getCustomData
function THStorageTransfer.registerXMLPaths(xmlSchema, xmlPath, ...)
    local function appendFunc(...)
        THUtils.call(function()
            THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#ratePerMinute", "Storage tranfer liter per in-game minute")
        end)
        return ...
    end
    return appendFunc(THStorageTransfer:superClass().registerXMLPaths(xmlSchema, xmlPath, ...))
end
function THStorageTransfer.new(isServer, isClient, customMt, target)
    customMt = customMt or THStorageTransfer_mt
    local self = THAnimatedProcessor.new(isServer, isClient, customMt, target)
    if self ~= nil then
        local customData = getCustomData(self)
        customData.maxUpdateTime = 500
        customData.transferRate = 10000
        customData.statusIds = THStorageTransfer.STATUS_ID
        customData.statusXMLKeys = THStorageTransfer.STATUS_XML_KEY
        customData.statusKeyToIndex = THStorageTransfer.STATUS_XML_INDEX
        customData.inputStorages = {}
        customData.outputStorages = {}
        customData.statusText[THStorageTransfer.STATUS.INPUT_EMPTY] = thI18n:getText("thTransferStatus_inputEmpty")
        customData.statusText[THStorageTransfer.STATUS.OUTPUT_FULL] = thI18n:getText("thTransferStatus_outputFull")
        customData.activateText = thI18n:getText("thAction_startTransfer")
        customData.deactivateText = thI18n:getText("thAction_stopTransfer")
        return self
    end
end
function THStorageTransfer:load(xmlFile, xmlKey, rootNode, customEnv, baseDirectory, i3dMappings, ...)
    local customData = getCustomData(self)
    if customData.isLoadFinished then
        return true
    end
    local success = false
    if THStorageTransfer:superClass().load(self, xmlFile, xmlKey, rootNode, customEnv, baseDirectory, i3dMappings, ...)
        and self:addStatusProcessor(self.processTransferStorages, self)
    then
        local transferRate = THUtils.getXMLValue(xmlFile, xmlKey, "#ratePerMinute", customData.transferRate)
        if transferRate < 0 then
            THUtils.xmlMsg(xmlKey, "ratePerMinute", transferRate, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
        else
            customData.transferRate = transferRate
            success = true
        end
    end
    customData.isEnabled = success == true
    return success
end
function THStorageTransfer:updateActionEvents(inputContext, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        THUtils.call(function()
            local actionEventId = customData.actionEventIds.toggleActive
            if actionEventId ~= nil then
                if next(customData.inputStorages) == nil
                    or next(customData.outputStorages) == nil
                then
                    thInputManager:setActionEventTextVisibility(actionEventId, false)
                    thInputManager:setActionEventActive(actionEventId, false)
                end
            end
        end)
    end
    return appendFunc(THStorageTransfer:superClass().updateActionEvents(self, inputContext, ...))
end
function THStorageTransfer:getInputStorages()
    local customData = getCustomData(self)
    return customData.inputStorages
end
function THStorageTransfer:getOutputStorages()
    local customData = getCustomData(self)
    return customData.outputStorages
end
function THStorageTransfer:getIsInputStorage(storage)
    local customData = getCustomData(self)
    if storage ~= nil and customData.inputStorages[storage] == true then
        return true
    end
    return false
end
function THStorageTransfer:getIsOutputStorage(storage)
    local customData = getCustomData(self)
    if storage ~= nil and customData.outputStorages[storage] == true then
        return true
    end
    return false
end
function THStorageTransfer:setInputStorages(storageList)
    local customData = getCustomData(self)
    THUtils.clearTable(customData.inputStorages)
    if THUtils.argIsValid(storageList == nil or type(storageList) == "table", "storageList", storageList)
        and self:getIsEnabled()
    then
        if THUtils.getIsType(storageList[1], Storage) then
            for storageIndex, storage in pairs(storageList) do
                if storageIndex == 1 or THUtils.getIsType(storage, Storage) then
                    customData.inputStorages[storage] = true
                end
            end
        else
            for storage in pairs(storageList) do
                if THUtils.getIsType(storage, Storage) then
                    customData.inputStorages[storage] = true
                end
            end
        end
    end
end
function THStorageTransfer:setOutputStorages(storageList)
    local customData = getCustomData(self)
    THUtils.clearTable(customData.outputStorages)
    if self:getIsEnabled()
        and THUtils.argIsValid(storageList == nil or type(storageList) == "table", "storageList", storageList)
    then
        if THUtils.getIsType(storageList[1], Storage) then
            for storageIndex, storage in pairs(storageList) do
                if storageIndex == 1 or THUtils.getIsType(storage, Storage) then
                    customData.outputStorages[storage] = true
                end
            end
        else
            for storage in pairs(storageList) do
                if THUtils.getIsType(storage, Storage) then
                    customData.outputStorages[storage] = true
                end
            end
        end
    end
end
function THStorageTransfer:processTransferStorages(_, elapsedTime)
    local customData = getCustomData(self)
    local mission = thMain.mission
    if self.isServer and mission ~= nil and mission.accessHandler ~= nil
        and self:getIsEnabled()
    then
        local currentStatus = self:getCurrentStatus()
        local lastStatus = currentStatus
        local elapsedMinutes = elapsedTime / 1000 / 60
        local effectFillType = FillType.UNKNOWN
        local isDebugEnabled = THDebugUtil.getIsDebugEnabled(debugFlagId, THDebugLevel.UPDATE)
        if customData.transferRate <= 0
            or next(customData.inputStorages) == nil
            or next(customData.outputStorages) == nil
        then
            currentStatus = THStorageTransfer.STATUS.DISABLED
        elseif currentStatus == THStorageTransfer.STATUS.DISABLED then
            currentStatus = THStorageTransfer.STATUS.STOPPED
        elseif currentStatus == THStorageTransfer.STATUS.RUNNING then
            if isDebugEnabled then
                THUtils.displayMsg("Storage transfer information:")
                THUtils.displayMsg("- Elapsed time: %s", elapsedMinutes)
            end
            if elapsedMinutes > 0 then
                local ownerFarmId = self:getOwnerFarmId()
                local transferAmount = customData.transferRate * elapsedMinutes
                local effectsAmount = 0
                local hasInputs = false
                for inputStorage in pairs(customData.inputStorages) do
                    local inputFillLevels = inputStorage:getFillLevels()
                    local thInputMoistureStorage = nil
                    if THMoistureStorage ~= nil then
                        thInputMoistureStorage = THMoistureStorage.getCustomData(inputStorage)
                    end
                    if inputFillLevels ~= nil
                        and (ownerFarmId == AccessHandler.EVERYONE or mission.accessHandler:canFarmAccess(ownerFarmId, inputStorage))
                    then
                        for fillTypeIndex, inputFillLevel in pairs(inputFillLevels) do
                            local totalAmount = math.min(inputFillLevel, transferAmount)
                            local amountRemaining = totalAmount
                            if inputFillLevel > 0.000001 and totalAmount > 0
                                and fillTypeIndex ~= FillType.UNKNOWN
                            then
                                local inputMoisture = nil
                                hasInputs = true
                                if totalAmount > effectsAmount then
                                    effectFillType = fillTypeIndex
                                    effectsAmount = totalAmount
                                end
                                if thInputMoistureStorage ~= nil then
                                    inputMoisture = thInputMoistureStorage:getFillTypeMoisture(fillTypeIndex)
                                end
                                for outputStorage in pairs(customData.outputStorages) do
                                    if ownerFarmId == AccessHandler.EVERYONE or mission.accessHandler:canFarmAccess(ownerFarmId, outputStorage) then
                                        local outputFreeCapacity = outputStorage:getFreeCapacity(fillTypeIndex)
                                        local thOutputMoistureStorage = nil
                                        if THMoistureStorage ~= nil and inputMoisture ~= nil then
                                            thOutputMoistureStorage = THMoistureStorage.getCustomData(outputStorage)
                                        end
                                        if outputFreeCapacity > 0 then
                                            local outputFillLevel = outputStorage:getFillLevel(fillTypeIndex)
                                            local newOutputFillLevel = outputFillLevel + math.min(amountRemaining, outputFreeCapacity)
                                            if thOutputMoistureStorage ~= nil and inputMoisture ~= nil then
                                                thOutputMoistureStorage:setFillTypeInputMoisture(fillTypeIndex, inputMoisture)
                                            end
                                            outputStorage:setFillLevel(newOutputFillLevel, fillTypeIndex)
                                            newOutputFillLevel = outputStorage:getFillLevel(fillTypeIndex)
                                            if newOutputFillLevel > outputFillLevel then
                                                local newInputFillLevel = math.max(0, inputFillLevel - (newOutputFillLevel - outputFillLevel))
                                                inputStorage:setFillLevel(newInputFillLevel, fillTypeIndex)
                                                newInputFillLevel = inputStorage:getFillLevel(fillTypeIndex)
                                                if newInputFillLevel < inputFillLevel then
                                                    amountRemaining = amountRemaining - (inputFillLevel - newInputFillLevel)
                                                    if isDebugEnabled then
                                                        local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
                                                        local amountMoved = inputFillLevel - newInputFillLevel
                                                        local inputStorageData = THPlaceable.getObjectData(inputStorage)
                                                        local outputStorageData = THPlaceable.getObjectData(outputStorage)
                                                        if inputStorageData ~= nil and outputStorageData ~= nil
                                                            and inputStorageData.id ~= nil and inputStorageData.id ~= ""
                                                            and outputStorageData.id ~= nil and outputStorageData.id ~= ""
                                                        then
                                                            THUtils.displayMsg("- Moved %s %q [%s] from %q to %q", amountMoved, fillTypeName, fillTypeIndex, inputStorageData.id, outputStorageData.id)
                                                        else
                                                            THUtils.displayMsg("- Moved %s %q [%s]", amountMoved, fillTypeName, fillTypeIndex)
                                                        end
                                                    end
                                                    if amountRemaining <= 0 then
                                                        break
                                                    end
                                                end
                                            end
                                        end
                                    end
                                end
                                if isDebugEnabled then
                                    THUtils.displayMsg("- Amount remaining: %s", amountRemaining)
                                end
                                if amountRemaining >= 0.000001 then
                                    currentStatus = THStorageTransfer.STATUS.OUTPUT_FULL
                                    break
                                end
                            end
                        end
                        if currentStatus ~= THStorageTransfer.STATUS.RUNNING then
                            break
                        end
                    end
                end
                if not hasInputs then
                    currentStatus = THStorageTransfer.STATUS.INPUT_EMPTY
                end
            end
        end
        if currentStatus ~= lastStatus then
            self:setCurrentStatus(currentStatus)
        end
        if effectFillType ~= customData.effectFillType then
            self:setEffectsFillType(effectFillType)
        end
    end
end