-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thStateControlSystem
local thFillTypeManager = thMain.thFillTypeManager
local thInputManager = thMain.thInputManager
local thMessageCenter = thMain.messageCenter
local thI18n = thMain.i18n
THPlaceableStateControl = {}
local debugFlagId = THDebugUtil.createDebugFlag("THPlaceableStateControl")
THPlaceableStateControl.debugFlagId = debugFlagId
THPlaceableStateControl.MOD_NAME = thMain.modName
THPlaceableStateControl.SPEC_NAME = thMain.dataKey
THPlaceableStateControl.STATE_BITS = 8 -- 255 unique states
THPlaceableStateControl.STATE_TYPE = {
    INPUT = 1,
    OUTPUT = 2
}
THPlaceableStateControl.STATE_TYPE_ID = {}
THPlaceableStateControl.MAX_STATE_TYPE = THPlaceableStateControl.STATE_TYPE.OUTPUT
THPlaceableStateControl.STATE_TYPE_BITS = 2
for typeId, typeIndex in pairs(THPlaceableStateControl.STATE_TYPE) do
    THPlaceableStateControl.STATE_TYPE_ID[typeIndex] = typeId
end
THPlaceableStateControl.HUD_TEXT = {
    INPUT_STATE = thI18n:getText("thInfoHud_inputState"),
    OUTPUT_STATE = thI18n:getText("thInfoHud_outputState"),
    INPUT_MOISTURE = thI18n:getText("thInfoHud_inputMoisture"),
    OUTPUT_MOISTURE = thI18n:getText("thInfoHud_outputMoisture"),
    DRYER_MOISTURE = thI18n:getText("thInfoHud_dryerMoisture"),
    STORAGE_EMPTY = thI18n:getText("infohud_siloEmpty"),
    TRANSFER_STATUS = thI18n:getText("thInfoHud_transferStatus")
}
THPlaceableStateControl.ACTION_TEXT = {
    SELECT_INPUT_STATE = thI18n:getText("thAction_selectInputState"),
    SELECT_OUTPUT_STATE = thI18n:getText("thAction_selectOutputState")
}
THPlaceableStateControl.BUFFER_UPDATE_TIME = THPlaceable.BUFFER_UPDATE_TIME
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THPlaceableStateControl.SPEC_NAME, thMain.modName, verbose)
end
THPlaceableStateControl.getSpecTable = getSpecTable
THSelectControlStateEvent = {}
local THSelectControlStateEvent_mt = THUtils.createClass(THSelectControlStateEvent, Event)
InitEventClass(THSelectControlStateEvent, "THSelectControlStateEvent")
function THPlaceableStateControl.initSpecialization()
end
function THPlaceableStateControl.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THPlaceable, specializations)
        and SpecializationUtil.hasSpecialization(PlaceableInfoTrigger, specializations)
end
function THPlaceableStateControl.registerXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THStateControlSystem")
    local xmlSubPath = THPlaceable.getConfigSchemaPath(THPlaceableStateControl, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.NODE_INDEX, xmlSubPath, "#triggerNode", "Controller trigger node index")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, xmlSubPath, "#isProduction", "Controller is a production point")
    local stationIdsPath = xmlSubPath .. ".stationIds"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, stationIdsPath, "#unloading", "Main unloading station id")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, stationIdsPath, "#loading", "Main loading station id")
    local controlStatesPath = xmlSubPath .. ".controlStates"
    local controlStatePath = controlStatesPath .. ".state(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, controlStatePath, "#id", "Control state unique id")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, controlStatePath, "#title", "Control state title")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, controlStatePath, "#extensionRange", "Silo extension max range")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, controlStatePath, ".storages", "Linked storage ids")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, controlStatePath, ".stations", "Linked station ids")
    THStatusFXHandler.registerXMLPaths(xmlSchema, xmlSubPath)
    THStatusFXHandler.registerStatusXMLPaths(xmlSchema, controlStatePath)
    for _, fxTypeIndex in pairs(THFXHandler.FX_TYPE) do
        local fxBasePath = controlStatePath .. "." .. THFXHandler.XML_BASE_KEY[fxTypeIndex]
        local fxSubPath = fxBasePath .. "." .. THFXHandler.FX_NAME[fxTypeIndex] .. "(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fxSubPath, "#linkedStates", "Linked control state ids")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, fxSubPath, "#isInput", "FX is active when this state is selected as input")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, fxSubPath, "#isOutput", "FX is active when this state is selected as output")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, fxSubPath, "#activeOnInput", "FX is active when specified input states are selected")
    end
    THStorageTransfer.registerXMLPaths(xmlSchema, xmlSubPath .. ".storageTransfer")
    local storageDryerPath = xmlSubPath .. ".storageDryer"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, storageDryerPath, ".storages", "Storages linked to the dryer")
    local productionPointPath = xmlSubPath .. ".productionPoint"
    ProductionPoint.registerXMLPaths(xmlSchema, productionPointPath)
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableStateControl.registerSavegameXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THStateControlSystem")
    local xmlSubPath = THPlaceable.getSavegameSchemaPath(THPlaceableStateControl, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, xmlSubPath, "#inputState", "Current input control state")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, xmlSubPath, "#outputState", "Current output control state")
    THStorageTransfer.registerSavegameXMLPaths(xmlSchema, xmlSubPath .. ".storageTransfer")
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableStateControl.registerEvents(placeableType)
    SpecializationUtil.registerEvent(placeableType, "thAA_onStateControlTransferStatusChanged")
end
function THPlaceableStateControl.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onPreLoad", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onLoad", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onDelete", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onUpdate", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onUpdateTick", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onReadStream", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onWriteStream", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onReadUpdateStream", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "onWriteUpdateStream", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "thAA_onEnterActionTrigger", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "thAA_onLeaveActionTrigger", THPlaceableStateControl)
    SpecializationUtil.registerEventListener(placeableType, "thAA_onStateControlTransferStatusChanged", THPlaceableStateControl)
end
function THPlaceableStateControl.registerFunctions(placeableType)
    SpecializationUtil.registerFunction(placeableType, "thAA_setupStateControl", THPlaceableStateControl.thAA_setupStateControl)
    SpecializationUtil.registerFunction(placeableType, "thAA_loadControlStates", THPlaceableStateControl.thAA_loadControlStates)
    SpecializationUtil.registerFunction(placeableType, "thAA_loadControlStateStorages", THPlaceableStateControl.thAA_loadControlStateStorages)
    SpecializationUtil.registerFunction(placeableType, "thAA_loadControlStateStations", THPlaceableStateControl.thAA_loadControlStateStations)
    SpecializationUtil.registerFunction(placeableType, "thAA_loadControlStateFX", THPlaceableStateControl.thAA_loadControlStateFX)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateStateControlActionEvents", THPlaceableStateControl.thAA_updateStateControlActionEvents)
    SpecializationUtil.registerFunction(placeableType, "thAA_readStateControlDataFromStream", THPlaceableStateControl.thAA_readStateControlDataFromStream)
    SpecializationUtil.registerFunction(placeableType, "thAA_writeStateControlDataToStream", THPlaceableStateControl.thAA_writeStateControlDataToStream)
    SpecializationUtil.registerFunction(placeableType, "thAA_getIsStateControlEnabled", THPlaceableStateControl.thAA_getIsStateControlEnabled)
    SpecializationUtil.registerFunction(placeableType, "thAA_getControlState", THPlaceableStateControl.thAA_getControlState)
    SpecializationUtil.registerFunction(placeableType, "thAA_getControlStateList", THPlaceableStateControl.thAA_getControlStateList)
    SpecializationUtil.registerFunction(placeableType, "thAA_getControlStateFillLevel", THPlaceableStateControl.thAA_getControlStateFillLevel)
    SpecializationUtil.registerFunction(placeableType, "thAA_getCurrentControlState", THPlaceableStateControl.thAA_getCurrentControlState)
    SpecializationUtil.registerFunction(placeableType, "thAA_selectControlState", THPlaceableStateControl.thAA_selectControlState)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateStateControlFX", THPlaceableStateControl.thAA_updateStateControlFX)
    SpecializationUtil.registerFunction(placeableType, "thAA_getControlStateStorages", THPlaceableStateControl.thAA_getControlStateStorages)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateStateControlStorages", THPlaceableStateControl.thAA_updateStateControlStorages)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateControlStateStorageBuffer", THPlaceableStateControl.thAA_updateControlStateStorageBuffer)
end
function THPlaceableStateControl.registerOverwrittenFunctions(placeableType)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "setOwnerFarmId", THPlaceableStateControl.setOwnerFarmId)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "collectPickObjects", THPlaceableStateControl.collectPickObjects)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "thAA_updateActionEvents", THPlaceableStateControl.thAA_updateActionEvents)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "thAA_updateExtendedStorageBuffer", THPlaceableStateControl.thAA_updateExtendedStorageBuffer)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "thAA_updateInfoHudEntries", THPlaceableStateControl.thAA_updateInfoHudEntries)
    THUtils.call(function()
        if THPlaceableMoisture ~= nil
            and SpecializationUtil.hasSpecialization(THPlaceableMoisture, placeableType.specializations)
        then
            SpecializationUtil.registerOverwrittenFunction(placeableType, "thAA_updateMoistureStorageBuffer", THPlaceableStateControl.thAA_updateMoistureStorageBuffer)
            SpecializationUtil.registerOverwrittenFunction(placeableType, "thAA_addMoistureStorageInfoHudEntries", THPlaceableStateControl.thAA_addMoistureStorageInfoHudEntries)
        end
    end)
end
function THPlaceableStateControl:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    local xmlFile = self.xmlFile
    local xmlSubKey = THPlaceable.getConfigXMLKey(THPlaceableStateControl)
    specTable.controlStates = {
        byId = {},
        byIndex = {}
    }
    specTable.currentControlState = {}
    specTable.defaultStorages = {}
    specTable.storageBuffer = {}
    specTable.storageBufferMapping = {}
    specTable.moistureStorageBuffer = {}
    specTable.dryerStorages = {}
    specTable.dryerStorageIds = {}
    specTable.dryerMoistureBuffer = {}
    specTable.activeStorages = {}
    specTable.activeStoragesArray = {}
    specTable.activeStations = {}
    specTable.activeStationsArray = {}
    specTable.isProduction = false
    specTable.isProductionCustom = false
    specTable.isProductionChanged = false
    specTable.isProductionLoaded = false
    specTable.productionDefaults = {}
    specTable.actionEventIds = {}
    specTable.placeableType = THPlaceableType.UNKNOWN
    specTable.isLoadFinished = false
    specTable.isPlacementFinished = false
    specTable.areControlStatesLoaded = false
    specTable.isStorageChanged = false
    specTable.canActivateTransfer = false
    specTable.bufferUpdateTime = 0
    specTable.dirtyFlag = self:getNextDirtyFlag()
    for _, stateTypeIndex in pairs(THPlaceableStateControl.STATE_TYPE) do
        specTable.currentControlState[stateTypeIndex] = 0
        specTable.defaultStorages[stateTypeIndex] = {}
        specTable.storageBuffer[stateTypeIndex] = {}
        specTable.moistureStorageBuffer[stateTypeIndex] = {}
    end
    specTable.isFXHandlerDirty = false
    specTable.functionParameters = {
        placeable = self
    }
    THUtils.call(function()
        if xmlFile ~= nil then
            specTable.isProduction = THUtils.getXMLValue(xmlFile, xmlSubKey, "#isProduction", false)
            specTable.isProductionCustom = xmlFile:hasProperty(xmlSubKey .. ".productionPoint") == true
            if specTable.isProductionCustom then
                if not specTable.isProductionChanged then
                    specTable.isProductionChanged = true
                end
            end
        end
        if not specTable.areHooksCreated then
            specTable.areHooksCreated = true
        end
    end)
end
function THPlaceableStateControl:onLoad(savegame)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        local customEnv = self.customEnvironment
        local components = self.components
        local i3dMappings = self.i3dMappings
        local baseDirectory = self.baseDirectory
        local xmlFile = self.xmlFile
        local xmlSubKey = THPlaceable.getConfigXMLKey(THPlaceableStateControl)
        if specTable.isLoadFinished then
            return
        end
        specTable.isControllerEnabled = false
        if xmlFile ~= nil and xmlFile:hasProperty(xmlSubKey)
            and self:thAA_setupStateControl(xmlFile, xmlSubKey, components, customEnv, baseDirectory, i3dMappings)
            and self:thAA_loadControlStates(xmlFile, xmlSubKey, components, customEnv, baseDirectory, i3dMappings)
        then
            local controlStatesArray, numControlStates = self:thAA_getControlStateList()
            if controlStatesArray == nil or numControlStates <= 0 then
                THPlaceable.errorMsg(self, false, "No valid control states defined")
            else
                if specTable.storageTransfer == nil then
                    local storageTransfer = THStorageTransfer.new(self.isServer, self.isClient, nil, self)
                    local transferData = THStorageTransfer.getCustomData(storageTransfer)
                    if storageTransfer ~= nil and transferData ~= nil then
                        local storageTransferKey = xmlSubKey .. ".storageTransfer"
                        if storageTransfer:load(xmlFile, storageTransferKey, components, customEnv, baseDirectory, i3dMappings) then
                            local function onTransferStatusChanged(...)
                                SpecializationUtil.raiseEvent(self, "thAA_onStateControlTransferStatusChanged", ...)
                            end
                            if storageTransfer:addStatusChangedListener(onTransferStatusChanged) then
                                storageTransfer:register(true)
                                specTable.storageTransfer = storageTransfer
                            end
                        end
                        if specTable.storageTransfer == nil then
                            storageTransfer:delete()
                            THPlaceable.errorMsg(self, true, "Could not initialize storage transfer")
                        else
                            if self.isClient and specTable.fxHandler == nil then
                                local fxHandler = THStatusFXHandler.new(self, transferData.statusXMLKeys)
                                if fxHandler ~= nil then
                                    if fxHandler:load(nil, xmlFile, xmlSubKey, components, customEnv, baseDirectory, i3dMappings) then
                                        local function hook_validateStatusFXObject(pSuperFunc, pFXHandler, pStatusIndex, pFXTypeIndex, pFXObjectIndex, pFXTableIndex, ...)
                                            local function appendFunc2(rIsActive, ...)
                                                if rIsActive then
                                                    THUtils.call(function()
                                                        local statusFXObject = pFXHandler:getStatusFXObject(pFXTypeIndex, pFXObjectIndex, pFXTableIndex)
                                                        if statusFXObject ~= nil then
                                                            local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
                                                            local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
                                                            if statusFXObject.inputStateIndex ~= nil or statusFXObject.outputStateIndex ~= nil then
                                                                rIsActive = false
                                                                if statusFXObject.inputStateIndex ~= nil
                                                                    and inputState ~= nil and statusFXObject.inputStateIndex == inputState.index
                                                                then
                                                                    rIsActive = true
                                                                elseif statusFXObject.outputStateIndex ~= nil
                                                                    and outputState ~= nil and statusFXObject.outputStateIndex == outputState.index
                                                                then
                                                                    rIsActive = true
                                                                end
                                                                if rIsActive then
                                                                    if statusFXObject.hasLinkedStates then
                                                                        local isLinkedStateFound = false
                                                                        for linkedStateIndex in pairs(statusFXObject.linkedStates) do
                                                                            if (inputState ~= nil and linkedStateIndex == inputState.index)
                                                                                or (outputState ~= nil and linkedStateIndex == outputState.index)
                                                                            then
                                                                                isLinkedStateFound = true
                                                                                break
                                                                            end
                                                                        end
                                                                        if isLinkedStateFound then
                                                                            if not statusFXObject.isActiveOnInputState then
                                                                                rIsActive = false
                                                                            end
                                                                        else
                                                                            if statusFXObject.isActiveOnInputState then
                                                                                rIsActive = false
                                                                            end
                                                                        end
                                                                    end
                                                                end
                                                            end
                                                        end
                                                    end)
                                                end
                                                return rIsActive, ...
                                            end
                                            return appendFunc2(pSuperFunc(pFXHandler, pStatusIndex, pFXTypeIndex, pFXObjectIndex, pFXTableIndex, ...))
                                        end
                                        THUtils.setFunctionHook(fxHandler, "validateStatusFXObject", nil, hook_validateStatusFXObject, false)
                                        specTable.fxObjectsIndex = fxHandler:loadStatusFXObjects(xmlFile, xmlSubKey)
                                        specTable.fxHandler = fxHandler
                                    else
                                        fxHandler:delete()
                                    end
                                end
                            end
                        end
                        if specTable.fxHandler == nil then
                            THPlaceable.warningMsg(self, "Failed to initialize fx handler")
                        else
                            for _, controlState in pairs(controlStatesArray) do
                                self:thAA_loadControlStateFX(controlState, xmlFile, xmlSubKey)
                            end
                        end
                        specTable.isControllerEnabled = true
                    end
                end
            end
        end
        local function onStorageAdded(pSelf, pOtherStorage, pOtherStation)
            if pSelf == self and specTable ~= nil
                and pOtherStorage ~= nil and pOtherStation ~= nil
                and pOtherStation ~= specTable.loadingStation
                and pOtherStation ~= specTable.unloadingStation
            then
                THUtils.call(function()
                    local thStationData = THPlaceable.getObjectData(pOtherStation)
                    if thStationData ~= nil and thStationData.loadingPlaceable == pSelf then
                        specTable.isStorageChanged = true
                        pSelf:raiseActive()
                    end
                end)
            end
        end
        local function onStorageRemoved(pSelf, pOtherStorage, pOtherStation)
            if pSelf == self and specTable ~= nil
                and pOtherStorage ~= nil and pOtherStation ~= nil
                and pOtherStation ~= specTable.loadingStation
                and pOtherStation ~= specTable.unloadingStation
            then
                THUtils.call(function()
                    local thStationData = THPlaceable.getObjectData(pOtherStation)
                    if thStationData ~= nil and thStationData.loadingPlaceable == pSelf then
                        specTable.isStorageChanged = true
                        pSelf:raiseActive()
                    end
                end)
            end
        end
        thMessageCenter:subscribe(MessageType.STORAGE_ADDED_TO_LOADING_STATION, onStorageAdded, self)
        thMessageCenter:subscribe(MessageType.STORAGE_ADDED_TO_UNLOADING_STATION, onStorageAdded, self)
        thMessageCenter:subscribe(MessageType.STORAGE_REMOVED_FROM_LOADING_STATION, onStorageRemoved, self)
        thMessageCenter:subscribe(MessageType.STORAGE_REMOVED_FROM_UNLOADING_STATION, onStorageRemoved, self)
        specTable.isLoadFinished = true
    end)
end
function THPlaceableStateControl:loadFromXMLFile(xmlFile, xmlKey)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        local xmlSubKey = THPlaceable.getSavegameXMLKey(THPlaceableStateControl, xmlKey)
        if self.isServer and xmlSubKey ~= "" then
            if self:thAA_getIsStateControlEnabled(true) then
                local inputStateIndex = THUtils.getXMLValue(xmlFile, xmlSubKey, "#inputState", 0)
                local outputStateIndex = THUtils.getXMLValue(xmlFile, xmlSubKey, "#outputState", 0)
                if inputStateIndex > 0 then
                    self:thAA_selectControlState(THPlaceableStateControl.STATE_TYPE.INPUT, inputStateIndex, true)
                end
                if outputStateIndex > 0 then
                    self:thAA_selectControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT, outputStateIndex, true)
                end
                if specTable.storageTransfer ~= nil then
                    specTable.storageTransfer:loadFromSavegameXML(xmlFile, xmlSubKey .. ".storageTransfer")
                end
            end
        end
    end)
end
function THPlaceableStateControl:saveToXMLFile(xmlFile, xmlKey, usedModNames)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        local xmlSubKey = THPlaceable.getSavegameXMLKey(THPlaceableStateControl, xmlKey)
        if self.isServer and xmlSubKey ~= ""
            and self:thAA_getIsStateControlEnabled()
        then
            local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
            local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
            if inputState ~= nil then
                xmlFile:setValue(xmlSubKey .. "#inputState", inputState.index)
            end
            if outputState ~= nil then
                xmlFile:setValue(xmlSubKey .. "#outputState", outputState.index)
            end
            if specTable.storageTransfer ~= nil then
                specTable.storageTransfer:saveToSavegameXML(xmlFile, xmlSubKey .. ".storageTransfer")
            end
        end
    end)
end
function THPlaceableStateControl:onFinalizePlacement()
    local specTable = getSpecTable(self)
    THUtils.call(function()
        local ownerFarmId = self:getOwnerFarmId()
        if self:thAA_getIsStateControlEnabled(true) then
            local unloadingStation = specTable.unloadingStation
            local loadingStation = specTable.loadingStation
            if unloadingStation ~= nil
                and unloadingStation.targetStorages ~= nil
            then
                for _, storage in pairs(unloadingStation.targetStorages) do
                    local storageData = THPlaceable.getObjectData(storage)
                    if storageData ~= nil and storageData.loadingPlaceable == self then
                        specTable.defaultStorages[THPlaceableStateControl.STATE_TYPE.INPUT][storage] = storage
                    end
                end
            end
            if loadingStation ~= nil
                and loadingStation.sourceStorages ~= nil
            then
                for _, storage in pairs(loadingStation.sourceStorages) do
                    local storageData = THPlaceable.getObjectData(storage)
                    if storageData ~= nil and storageData.loadingPlaceable == self then
                        specTable.defaultStorages[THPlaceableStateControl.STATE_TYPE.OUTPUT][storage] = storage
                    end
                end
            end
            local lastTransferStatus = nil
            if specTable.storageTransfer ~= nil then
                lastTransferStatus = specTable.storageTransfer:getCurrentStatus()
                specTable.storageTransfer:setOwnerFarmId(ownerFarmId)
            end
            specTable.isPlacementFinished = true
            for _, stateTypeIndex in pairs(THPlaceableStateControl.STATE_TYPE) do
                local controlStateIndex = specTable.currentControlState[stateTypeIndex] or 0
                if controlStateIndex > 0 then
                    self:thAA_selectControlState(stateTypeIndex, controlStateIndex, true)
                end
            end
            if specTable.storageTransfer ~= nil
                and lastTransferStatus ~= nil
                and lastTransferStatus ~= THStorageTransfer.STATUS.DISABLED
                and lastTransferStatus ~= THStorageTransfer.STATUS.STOPPED
            then
                specTable.storageTransfer:setCurrentStatus(lastTransferStatus)
            end
            if THPlaceableMoisture ~= nil then
                local thMoistureSpec = THPlaceableMoisture.getSpecTable(self)
                if thMoistureSpec ~= nil and thMoistureSpec.storageDryer ~= nil then
                    THUtils.clearTable(specTable.dryerStorages)
                    for dryerStorageId in pairs(specTable.dryerStorageIds) do
                        local storageList = self:thAA_getRegisteredObjectsById(dryerStorageId, THPlaceable.OBJECT_TYPE.STORAGE)
                        if storageList == nil or next(storageList) == nil then
                            THPlaceable.errorMsg(self, false, "Cannot find dryer storage id %q", dryerStorageId)
                            specTable.dryerStorageIds[dryerStorageId] = nil
                        else
                            for storage in pairs(storageList) do
                                specTable.dryerStorages[storage] = storage
                            end
                        end
                    end
                    local dryerStatus = thMoistureSpec.storageDryer:getCurrentStatus()
                    thMoistureSpec.storageDryer:setOutputStorages(specTable.dryerStorages)
                    if dryerStatus ~= nil
                        and dryerStatus ~= THStorageDryer.STATUS.DISABLED
                        and dryerStatus ~= THStorageDryer.STATUS.STOPPED
                    then
                        thMoistureSpec.storageDryer:setCurrentStatus(dryerStatus)
                    end
                end
            end
            self:raiseActive()
            if self.isServer then
                self:raiseDirtyFlags(specTable.dirtyFlag)
            end
            if THDebugUtil.getIsDebugEnabled(debugFlagId) then
                THPlaceable.displayMsg(self, "THPlaceableStateControl information:")
                THUtils.printTable(specTable, 1)
                THPlaceable.displayMsg(self, "- Control states:")
                THUtils.printTable(specTable.controlStates.byIndex, 2)
                THPlaceable.displayMsg(self, "- Control state ids:")
                THUtils.printTable(specTable.controlStates.byId)
                if unloadingStation ~= nil then
                    THPlaceable.displayMsg(self, "- Default unloading station:")
                    THUtils.printTable(unloadingStation)
                    if unloadingStation.targetStorages ~= nil then
                        THPlaceable.displayMsg(self, "- Unloading station target storages:")
                        THUtils.printTable(unloadingStation.targetStorages)
                    end
                end
                if loadingStation ~= nil then
                    THPlaceable.displayMsg(self, "- Default loading station:")
                    THUtils.printTable(loadingStation)
                    if loadingStation.sourceStorages ~= nil then
                        THPlaceable.displayMsg(self, "- Loading station source storages:")
                        THUtils.printTable(loadingStation.sourceStorages)
                    end
                end
                THPlaceable.displayMsg(self, "- Default storages:")
                THUtils.printTable(specTable.defaultStorages, 1)
                if specTable.fxHandler ~= nil then
                    THPlaceable.displayMsg(self, "- FX information:")
                    THUtils.printTable(specTable.fxHandler.fxObjectGroups, 4)
                end
                local transferData = THStorageTransfer.getCustomData(specTable.storageTransfer)
                if transferData ~= nil then
                    THPlaceable.displayMsg(self, "THStorageTransfer information:")
                    THUtils.printTable(transferData, 1)
                    if transferData.fxHandler ~= nil then
                        THPlaceable.displayMsg(self, "- FX information:")
                        THUtils.printTable(transferData.fxHandler.fxObjectGroups, 4)
                    end
                end
            end
        end
    end)
end
function THPlaceableStateControl:onDelete()
    local specTable = getSpecTable(self)
    if specTable ~= nil then
        THUtils.call(function()
            if specTable.controlTrigger ~= nil then
                removeTrigger(specTable.controlTrigger)
                specTable.controlTrigger = nil
            end
            if specTable.storageTransfer ~= nil then
                specTable.storageTransfer:delete()
                specTable.storageTransfer = nil
            end
            if specTable.fxHandler ~= nil then
                specTable.fxHandler:delete()
                specTable.fxHandler = nil
            end
        end)
    end
end
function THPlaceableStateControl:onUpdate(dt)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        if self:thAA_getIsStateControlEnabled() then
            if self.isClient then
                if specTable.bufferUpdateTime >= THPlaceableStateControl.BUFFER_UPDATE_TIME then
                    self:thAA_updateControlStateStorageBuffer()
                    specTable.bufferUpdateTime = 0
                else
                    specTable.bufferUpdateTime = specTable.bufferUpdateTime + dt
                    self:raiseActive()
                end
            end
            if specTable.fxHandler ~= nil then
                if specTable.isFXHandlerDirty then
                    specTable.fxHandler:updateActiveFXObjects()
                    specTable.isFXHandlerDirty = false
                end
            end
        end
    end)
end
function THPlaceableStateControl:onUpdateTick(dt)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        if specTable.isStorageChanged then
            self:thAA_updateStateControlStorages()
            specTable.isStorageChanged = false
        end
    end)
end
function THPlaceableStateControl:onReadStream(streamId, connection)
    local specTable = getSpecTable(self)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            if specTable.storageTransfer ~= nil then
                local storageTransferId = NetworkUtil.readNodeObjectId(streamId)
                specTable.storageTransfer:readStream(streamId, connection)
                g_client:finishRegisterObject(specTable.storageTransfer, storageTransferId)
            end
            specTable.isPlacementFinished = true
            self:thAA_readStateControlDataFromStream(streamId, connection)
        else
            specTable.isControllerEnabled = false
        end
    end
end
function THPlaceableStateControl:onWriteStream(streamId, connection)
    local specTable = getSpecTable(self)
    if not connection:getIsServer() then
        local isEnabled = self:thAA_getIsStateControlEnabled(true)
        if streamWriteBool(streamId, isEnabled == true) then
            if specTable.storageTransfer ~= nil then
                NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(specTable.storageTransfer))
                specTable.storageTransfer:writeStream(streamId, connection)
                g_server:registerObjectInStream(connection, specTable.storageTransfer)
            end
            self:thAA_writeStateControlDataToStream(streamId, connection)
        end
    end
end
function THPlaceableStateControl:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        self:thAA_readStateControlDataFromStream(streamId, connection, timestamp)
    end
end
function THPlaceableStateControl:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        self:thAA_writeStateControlDataToStream(streamId, connection, dirtyMask)
    end
end
function THPlaceableStateControl:thAA_onEnterActionTrigger(inputContext)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        if self:thAA_getIsStateControlEnabled()
            and self:thAA_getIsActiveForInput(inputContext, true)
        then
            local _, numControlStates = self:thAA_getControlStateList()
            local actionEventId = nil
            if numControlStates > 1 then
                _, actionEventId = thInputManager:registerActionEvent("THAA_PLAYER_FUNCTION", 1, specTable.actionEventIds.selectInputState, self, THPlaceableStateControl.thAA_actionSelectControlState, false, true, false, false, THPlaceableStateControl.STATE_TYPE.INPUT)
                if actionEventId ~= nil then
                    thInputManager:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
                    thInputManager:setActionEventText(actionEventId, THPlaceableStateControl.ACTION_TEXT.SELECT_INPUT_STATE)
                end
                specTable.actionEventIds.selectInputState = actionEventId
                _, actionEventId = thInputManager:registerActionEvent("THAA_PLAYER_FUNCTION", 1, specTable.actionEventIds.selectOutputState, self, THPlaceableStateControl.thAA_actionSelectControlState, false, true, false, false, THPlaceableStateControl.STATE_TYPE.OUTPUT)
                if actionEventId ~= nil then
                    thInputManager:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
                    thInputManager:setActionEventText(actionEventId, THPlaceableStateControl.ACTION_TEXT.SELECT_OUTPUT_STATE)
                end
                specTable.actionEventIds.selectOutputState = actionEventId
                if specTable.storageTransfer ~= nil then
                    specTable.storageTransfer:registerActionEvents(inputContext)
                end
                self:thAA_updateStateControlActionEvents(inputContext)
                self:raiseActive()
            end
        end
    end)
end
function THPlaceableStateControl:thAA_onLeaveActionTrigger(inputContext)
    local specTable = getSpecTable(self)
    if specTable ~= nil then
        THUtils.call(function()
            for eventName, eventId in pairs(specTable.actionEventIds) do
                thInputManager:removeActionEvent(eventId)
                specTable.actionEventIds[eventName] = nil
            end
            if specTable.storageTransfer ~= nil then
                specTable.storageTransfer:removeActionEvents(inputContext)
            end
        end)
    end
end
function THPlaceableStateControl:thAA_onStateControlTransferStatusChanged(transferData, currentStatus, lastStatus, ...)
    THUtils.call(function()
        self:thAA_updateStateControlFX()
        self:thAA_updateStateControlActionEvents()
        self:raiseActive()
    end)
end
function THPlaceableStateControl:thAA_actionSelectControlState(actionName, inputValue, stateTypeIndex)
    local specTable = getSpecTable(self)
    if specTable ~= nil and stateTypeIndex ~= nil then
        THUtils.call(function()
            self:thAA_selectControlState(stateTypeIndex)
        end)
    end
end
function THPlaceableStateControl:thAA_setupStateControl(xmlFile, xmlKey, components, customEnv, baseDirectory, i3dMappings)
    local specTable = getSpecTable(self)
    if specTable.isLoadFinished then
        return true
    end
    specTable.placeableType = THPlaceableType.UNKNOWN
    local loadingStation, unloadingStation = nil, nil
    local areStationsValid = false
    local stationIdsKey = xmlKey .. ".stationIds"
    local loadingStationId = THUtils.getXMLValue(xmlFile, stationIdsKey, "#loading", ""):upper()
    local unloadingStationId = THUtils.getXMLValue(xmlFile, stationIdsKey, "#unloading", ""):upper()
    if loadingStationId ~= "" then
        local otherStation = self:thAA_getRegisteredObject(loadingStationId, THPlaceable.OBJECT_TYPE.LOADING_STATION, nil, true)
        if otherStation == nil then
            THUtils.xmlMsg(stationIdsKey, "loading", loadingStationId, true, THXMLMessage.INVALID_VALUE)
        else
            loadingStation = otherStation
        end
    end
    if unloadingStationId ~= "" then
        local otherStation = self:thAA_getRegisteredObject(unloadingStationId, THPlaceable.OBJECT_TYPE.UNLOADING_STATION)
        if otherStation == nil then
            THUtils.xmlMsg(stationIdsKey, "unloading", unloadingStationId, true, THXMLMessage.INVALID_VALUE)
        else
            unloadingStation = otherStation
        end
    end
    if specTable.isProduction then
        local productionSpec = self.spec_productionPoint
        if productionSpec == nil or productionSpec.productionPoint == nil then
            THPlaceable.errorMsg(self, false, "Cannot find production point")
        else
            local productionPoint = productionSpec.productionPoint
            if loadingStation == nil and productionPoint.loadingStation == nil then
                THPlaceable.errorMsg(self, false, "Production needs a valid loading station to be used with state control system")
            elseif unloadingStation == nil and productionPoint.unloadingStation == nil then
                THPlaceable.errorMsg(self, false, "Production needs a valid unloading station to be used with state control system")
            elseif productionPoint.storage == nil then
                THPlaceable.errorMsg(self, false, "Production needs a valid storage to be used with state control system")
            else
                specTable.productionDefaults.unloadingStation = productionPoint.unloadingStation
                specTable.productionDefaults.loadingStation = productionPoint.loadingStation
                specTable.productionDefaults.storage = productionPoint.storage
                specTable.placeableType = THPlaceableType.PRODUCTION
                specTable.productionPoint = productionPoint
                loadingStation = loadingStation or productionPoint.loadingStation
                unloadingStation = unloadingStation or productionPoint.unloadingStation
                areStationsValid = true
            end
        end
    else
        local siloSpec = self.spec_silo
        if siloSpec ~= nil
            and (loadingStation ~= nil or siloSpec.loadingStation ~= nil)
            and (unloadingStation ~= nil or siloSpec.unloadingStation ~= nil)
        then
            specTable.placeableType = THPlaceableType.SILO
            loadingStation = loadingStation or siloSpec.loadingStation
            unloadingStation = unloadingStation or siloSpec.unloadingStation
            areStationsValid = true
        else
            if loadingStation == nil then
                local stationsArray, numStations = self:thAA_getExtendedObjects(THPlaceable.OBJECT_TYPE.LOADING_STATION)
                if stationsArray ~= nil and numStations > 0 then
                    loadingStation = stationsArray[1]
                else
                    THPlaceable.errorMsg(self, false, "Could not find default loading station")
                end
            end
            if unloadingStation == nil then
                local stationsArray, numStations = self:thAA_getExtendedObjects(THPlaceable.OBJECT_TYPE.UNLOADING_STATION)
                if stationsArray ~= nil and numStations > 0 then
                    unloadingStation = stationsArray[1]
                else
                    THPlaceable.errorMsg(self, false, "Could not find default unloading station")
                end
            end
            if loadingStation ~= nil and unloadingStation ~= nil then
                specTable.placeableType = THPlaceableType.UNKNOWN
                areStationsValid = true
            end
        end
    end
    if areStationsValid
        and unloadingStation ~= nil and loadingStation ~= nil
    then
        unloadingStation.storageRadius = 0
        unloadingStation.supportsExtension = false
        specTable.unloadingStation = unloadingStation
        loadingStation.storageRadius = 0
        loadingStation.supportsExtension = false
        specTable.loadingStation = loadingStation
        if THPlaceableMoisture ~= nil then
            local storageDryerKey = xmlKey .. ".storageDryer"
            local thMoistureSpec = THPlaceableMoisture.getSpecTable(self)
            if thMoistureSpec ~= nil and xmlFile:hasProperty(storageDryerKey) then
                local dryerStorageIds = THUtils.getXMLValue(xmlFile, storageDryerKey, ".storages", ""):upper()
                if dryerStorageIds ~= "" then
                    local dryerStorageIdList = THUtils.splitString(dryerStorageIds)
                    if dryerStorageIdList ~= nil then
                        for dryerStorageId in pairs(dryerStorageIdList) do
                            if not THUtils.validateId(dryerStorageId) then
                                THUtils.xmlMsg(storageDryerKey, "storage", dryerStorageId, true, THXMLMessage.INVALID_VALUE)
                            else
                                specTable.dryerStorageIds[dryerStorageId] = true
                            end
                        end
                    end
                end
            end
        end
        return true
    end
    return false
end
function THPlaceableStateControl:thAA_loadControlStates(xmlFile, xmlKey, components, customEnv, baseDirectory, i3dMappings)
    local specTable = getSpecTable(self)
    if specTable.isLoadFinished
        or specTable.areControlStatesLoaded
    then
        return true
    end
    local controlStatesKey = xmlKey .. ".controlStates"
    local success = true
    xmlFile:iterate(controlStatesKey .. ".state", function(_, pStateKey)
        local stateId = THUtils.getXMLValue(xmlFile, pStateKey, "#id", ""):upper()
        local stateTitle = THUtils.getXMLValue(xmlFile, pStateKey, "#title")
        local extensionRange = THUtils.getXMLValue(xmlFile, pStateKey, "#extensionRange", 0)
        if not THUtils.validateId(stateId) then
            THUtils.xmlMsg(pStateKey, "id", stateId, true, THXMLMessage.INVALID_VALUE)
        elseif specTable.controlStates.byId[stateId] ~= nil then
            THUtils.xmlMsg(pStateKey, "id", stateId, true, THXMLMessage.DUPLICATE_ENTRY)
        else
            local stateIndex = #specTable.controlStates.byIndex + 1
            local maxNumStates = (2 ^ THPlaceableStateControl.STATE_BITS) - 1
            if stateIndex > maxNumStates then
                THPlaceable.errorMsg(self, false, "Too many control states registered (%s)", maxNumStates)
                return false
            end
            if extensionRange < 0 then
                THUtils.xmlMsg(pStateKey, "extensionRange", extensionRange, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                extensionRange = 0
            end
            if stateTitle == nil then
                stateTitle = stateId
            elseif stateTitle ~= "" then
                stateTitle = thI18n:convertText(stateTitle, customEnv)
            end
            local controlState = {}
            controlState.id = stateId
            controlState.title = stateTitle
            controlState.index = stateIndex
            controlState.xmlKey = pStateKey
            controlState.linkedStorages = {}
            controlState.linkedStoragesArray = {}
            controlState.linkedStorageIds = {}
            controlState.linkedStations = {}
            controlState.linkedStationArray = {}
            controlState.linkedStationIds = {}
            controlState.extensionRange = extensionRange
            if self:thAA_loadControlStateStorages(controlState, xmlFile, pStateKey)
                and self:thAA_loadControlStateStations(controlState, xmlFile, pStateKey)
            then
                specTable.controlStates.byIndex[stateIndex] = controlState
                specTable.controlStates.byId[stateId] = controlState
                for _, stateTypeIndex in pairs(THPlaceableStateControl.STATE_TYPE) do
                    if specTable.currentControlState[stateTypeIndex] == nil
                        or specTable.currentControlState[stateTypeIndex] <= 0
                    then
                        specTable.currentControlState[stateTypeIndex] = 1
                    end
                end
            end
        end
    end)
    specTable.areControlStatesLoaded = true
    return success
end
function THPlaceableStateControl:thAA_loadControlStateStorages(controlState, xmlFile, xmlKey)
    local specTable = getSpecTable(self)
    if specTable.isLoadFinished
        or specTable.areControlStatesLoaded
    then
        return true
    end
    local storageIds = THUtils.getXMLValue(xmlFile, xmlKey, ".storages", ""):upper()
    local success = true
    if storageIds ~= "" then
        local storageIdList = THUtils.splitString(storageIds)
        if storageIdList ~= nil then
            for storageId in pairs(storageIdList) do
                local storageList = self:thAA_getRegisteredObjectsById(storageId, THPlaceable.OBJECT_TYPE.STORAGE)
                if storageList == nil or next(storageList) == nil then
                    THUtils.xmlMsg(xmlKey, "storage", storageId, true, THXMLMessage.INVALID_VALUE)
                else
                    for storage in pairs(storageList) do
                        if not controlState.linkedStorages[storage] then
                            local linkedStorageIndex = #controlState.linkedStoragesArray + 1
                            controlState.linkedStoragesArray[linkedStorageIndex] = storage
                            controlState.linkedStorages[storage] = true
                        end
                    end
                    controlState.linkedStorageIds[storageId] = true
                end
            end
        end
    end
    return success
end
function THPlaceableStateControl:thAA_loadControlStateStations(controlState, xmlFile, xmlKey)
    return true
end
function THPlaceableStateControl:thAA_loadControlStateFX(controlState, xmlFile, xmlKey)
    local specTable = getSpecTable(self)
    local fxHandler = specTable.fxHandler
    if specTable.isLoadFinished then
        return true
    end
    if fxHandler ~= nil
        and specTable.areControlStatesLoaded
        and controlState.fxObjectsIndex == nil
    then
        local fxObjectsIndex = fxHandler:loadStatusFXObjects(xmlFile, controlState.xmlKey)
        if fxObjectsIndex ~= nil and fxObjectsIndex > 0 then
            controlState.fxObjectsIndex = fxObjectsIndex
            for _, fxTypeIndex in pairs(THFXHandler.FX_TYPE) do
                local statusFXObjects = fxHandler:getStatusFXObjects(fxTypeIndex, controlState.fxObjectsIndex)
                if statusFXObjects ~= nil then
                    for _, statusFXObject in pairs(statusFXObjects) do
                        local objectXMLKey = statusFXObject.xmlKey
                        local linkedStateIds = THUtils.getXMLValue(xmlFile, objectXMLKey, "#linkedStates", ""):upper()
                        local isInputState = THUtils.getXMLValue(xmlFile, objectXMLKey, "#isInput", true)
                        local isOutputState = THUtils.getXMLValue(xmlFile, objectXMLKey, "#isOutput", true)
                        statusFXObject.inputStateIndex = 0
                        statusFXObject.outputStateIndex = 0
                        statusFXObject.linkedStates = {}
                        statusFXObject.hasLinkedStates = linkedStateIds ~= ""
                        statusFXObject.isActiveOnInputState = THUtils.getXMLValue(xmlFile, objectXMLKey, "#activeOnInput", true)
                        if isInputState or isOutputState then
                            if isInputState then
                                statusFXObject.inputStateIndex = controlState.index
                            end
                            if isOutputState then
                                statusFXObject.outputStateIndex = controlState.index
                            end
                        end
                        if statusFXObject.hasLinkedStates then
                            local linkedStateIdList = THUtils.splitString(linkedStateIds)
                            if linkedStateIdList ~= nil then
                                for linkedStateId in pairs(linkedStateIdList) do
                                    local linkedState = self:thAA_getControlState(linkedStateId)
                                    if linkedState ~= nil then
                                        statusFXObject.linkedStates[linkedState.index] = true
                                    else
                                        THUtils.xmlMsg(objectXMLKey, "linkedState", linkedStateId, true, THXMLMessage.INVALID_VALUE)
                                    end
                                end
                            end
                            if next(statusFXObject.linkedStates) == nil then
                                THUtils.xmlMsg(objectXMLKey, nil, nil, true, "no valid linked control states defined")
                            end
                        end
                    end
                end
            end
            return true
        end
    end
    return false
end
function THPlaceableStateControl:thAA_updateStateControlActionEvents(inputContext)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        local isTransferAllowed = false
        if self:thAA_getIsStateControlEnabled()
            and self:thAA_getIsActiveForInput(inputContext)
            and self:thAA_getHasDisplayInfo()
        then
            local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
            local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
            for _, eventId in pairs(specTable.actionEventIds) do
                local isEventActive = true
                if eventId == specTable.actionEventIds.selectInputState then
                    if inputState == nil then
                        isEventActive = false
                    end
                elseif eventId == specTable.actionEventIds.selectOutputState then
                    if outputState == nil then
                        isEventActive = false
                    end
                end
                thInputManager:setActionEventTextVisibility(eventId, isEventActive)
                thInputManager:setActionEventActive(eventId, isEventActive)
            end
            if inputState ~= nil and outputState ~= nil and inputState ~= outputState then
                isTransferAllowed = specTable.canActivateTransfer == true
            end
        else
            for _, eventId in pairs(specTable.actionEventIds) do
                thInputManager:setActionEventTextVisibility(eventId, false)
                thInputManager:setActionEventActive(eventId, false)
            end
        end
        if specTable.storageTransfer ~= nil then
            specTable.storageTransfer:setIsActivatable(isTransferAllowed)
        end
    end)
end
function THPlaceableStateControl:thAA_readStateControlDataFromStream(streamId, connection, timestamp)
    local specTable = getSpecTable(self)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isEnabled
            if streamReadBool(streamId) then -- dirtyFlag
                for stateTypeIndex = 1, THPlaceableStateControl.MAX_STATE_TYPE do
                    local controlStateIndex = streamReadUIntN(streamId, THPlaceableStateControl.STATE_BITS)
                    if controlStateIndex > 0 then
                        THUtils.call(function()
                            self:thAA_selectControlState(stateTypeIndex, controlStateIndex, true)
                        end)
                    end
                end
            end
        else
            specTable.isControllerEnabled = false
        end
    end
end
function THPlaceableStateControl:thAA_writeStateControlDataToStream(streamId, connection, dirtyMask)
    local specTable = getSpecTable(self)
    if not connection:getIsServer() then
        local isEnabled = specTable.isControllerEnabled == true
        if streamWriteBool(streamId, isEnabled) then
            if streamWriteBool(streamId, dirtyMask == nil or bitAND(dirtyMask, specTable.dirtyFlag) ~= 0) then
                for stateTypeIndex = 1, THPlaceableStateControl.MAX_STATE_TYPE do
                    local controlStateIndex = specTable.currentControlState[stateTypeIndex] or 0
                    streamWriteUIntN(streamId, controlStateIndex, THPlaceableStateControl.STATE_BITS)
                end
            end
        end
    end
end
function THPlaceableStateControl:thAA_getIsStateControlEnabled(ignorePlacementFinished)
    local specTable = getSpecTable(self)
    if THUtils.argIsValid(not ignorePlacementFinished or ignorePlacementFinished == true, "ignorePlacementFinished", ignorePlacementFinished) then
        if specTable.isControllerEnabled then
            if ignorePlacementFinished or specTable.isPlacementFinished then
                return true
            end
        end
    end
    return false
end
function THPlaceableStateControl:thAA_getControlState(stateId)
    local specTable = getSpecTable(self)
    local idType = type(stateId)
    if idType == "string" then
        return specTable.controlStates.byId[stateId:upper()]
    elseif idType == "number" then
        return specTable.controlStates.byIndex[stateId]
    elseif stateId ~= nil then
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "stateId", stateId)
    end
end
function THPlaceableStateControl:thAA_getControlStateList()
    local specTable = getSpecTable(self)
    return specTable.controlStates.byIndex, #specTable.controlStates.byIndex
end
function THPlaceableStateControl:thAA_getControlStateFillLevel(fillTypeIndex, stateTypeIndex)
    local specTable = getSpecTable(self)
    local storageList = specTable.activeStorages
    if stateTypeIndex == THPlaceableStateControl.STATE_TYPE.INPUT then
        if specTable.unloadingStation ~= nil then
            storageList = specTable.unloadingStation.targetStorages
        end
    elseif stateTypeIndex == THPlaceableStateControl.STATE_TYPE.OUTPUT then
        if specTable.loadingStation ~= nil then
            storageList = specTable.loadingStation.sourceStorages
        end
    elseif stateTypeIndex ~= nil then
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "stateTypeIndex", stateTypeIndex)
        return 0
    end
    if storageList ~= nil then
        local totalFillLevel = 0
        if fillTypeIndex == nil then
            for storage in pairs(storageList) do
                if storage.getFillLevels ~= nil then
                    local storageFillLevels = storage:getFillLevels()
                    if storageFillLevels ~= nil then
                        for storageFillType, storageFillLevel in pairs(storageFillLevels) do
                            if storageFillLevel > 0
                                and storageFillType ~= nil and storageFillType ~= FillType.UNKNOWN
                            then
                                totalFillLevel = totalFillLevel + storageFillLevel
                            end
                        end
                    end
                end
            end
        elseif fillTypeIndex ~= FillType.UNKNOWN then
            for storage in pairs(storageList) do
                if storage.getFillLevel ~= nil then
                    local fillLevel = storage:getFillLevel(fillTypeIndex) or 0
                    if fillLevel > 0 then
                        totalFillLevel = totalFillLevel + fillLevel
                    end
                end
            end
        end
        return totalFillLevel
    end
    return 0
end
function THPlaceableStateControl:thAA_getCurrentControlState(stateTypeIndex)
    local specTable = getSpecTable(self)
    if THPlaceableStateControl.STATE_TYPE_ID[stateTypeIndex] == nil then
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "stateTypeIndex", stateTypeIndex)
        return
    end
    local controlStateIndex = specTable.currentControlState[stateTypeIndex] or 0
    if controlStateIndex > 0 then
        return specTable.controlStates.byIndex[controlStateIndex]
    end
end
function THPlaceableStateControl:thAA_selectControlState(stateTypeIndex, stateId, noEventSend)
    local specTable = getSpecTable(self)
    local stateTypeId = THPlaceableStateControl.STATE_TYPE_ID[stateTypeIndex]
    if THUtils.argIsValid(stateTypeId ~= nil, "stateTypeIndex", stateTypeIndex)
        and THUtils.argIsValid(not noEventSend or noEventSend == true, "noEventSend", noEventSend)
        and self:thAA_getIsStateControlEnabled(true)
    then
        if self.isServer or noEventSend == true then
            local controlState = nil
            if stateId ~= nil then
                controlState = self:thAA_getControlState(stateId)
            else
                local numControlStates = #specTable.controlStates.byIndex
                if numControlStates > 0 then
                    local controlStateIndex = (specTable.currentControlState[stateTypeIndex] or 0) + 1
                    if controlStateIndex <= 0 or controlStateIndex > numControlStates then
                        controlStateIndex = 1
                    end
                    controlState = self:thAA_getControlState(controlStateIndex)
                end
            end
            if controlState ~= nil then
                local isControlStateChanged = false
                if specTable.currentControlState[stateTypeIndex] ~= controlState.index then
                    specTable.currentControlState[stateTypeIndex] = controlState.index
                    isControlStateChanged = true
                end
                if specTable.isPlacementFinished then
                    if THDebugUtil.getIsDebugEnabled(debugFlagId) then
                        THPlaceable.displayMsg(self, "Selected %q control state", stateTypeId)
                        THUtils.printTable(controlState)
                    end
                    if specTable.storageTransfer ~= nil then
                        if isControlStateChanged then
                            specTable.storageTransfer:setCurrentStatus(THStorageTransfer.STATUS.STOPPED, noEventSend)
                        end
                    end
                    self:thAA_updateStateControlStorages()
                    self:thAA_updateStateControlFX()
                    self:thAA_updateStateControlActionEvents()
                    self:raiseActive()
                end
                if self.isServer and noEventSend ~= true then
                    self:raiseDirtyFlags(specTable.dirtyFlag)
                end
                return controlState
            end
        elseif self.isClient and stateId == nil then
            local newEvent = THSelectControlStateEvent.new(self, stateTypeIndex)
            if newEvent ~= nil and newEvent.isEventValid then
                g_client:getServerConnection():sendEvent(newEvent)
            end
        end
    end
end
function THPlaceableStateControl:thAA_updateStateControlFX()
    local specTable = getSpecTable(self)
    local storageTransfer = specTable.storageTransfer
    local fxHandler = specTable.fxHandler
    if self:thAA_getIsStateControlEnabled()
        and fxHandler ~= nil
    then
        local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
        local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
        local transferStatus = THStorageTransfer.STATUS.DISABLED
        local isDebugEnabled = THDebugUtil.getIsDebugEnabled(debugFlagId)
        if storageTransfer ~= nil then
            transferStatus = storageTransfer:getCurrentStatus()
        end
        fxHandler:stopFX()
        if specTable.fxObjectsIndex ~= nil and specTable.fxObjectsIndex > 0 then
            fxHandler:startStatusFX(transferStatus, false, nil, specTable.fxObjectsIndex)
        end
        local function processStateFX(pOtherState)
            if pOtherState.fxObjectsIndex ~= nil and pOtherState.fxObjectsIndex > 0 then
                fxHandler:startStatusFX(transferStatus, false, nil, pOtherState.fxObjectsIndex)
                if isDebugEnabled then
                    local fxObjectsTable = fxHandler.statusFXObjects[pOtherState.fxObjectsIndex]
                    if fxObjectsTable ~= nil then
                        THPlaceable.displayMsg(self, "Status FX objects: %s", pOtherState.id)
                        THUtils.printTable(fxObjectsTable, 3)
                    end
                end
            end
        end
        if inputState ~= nil and inputState ~= outputState then
            processStateFX(inputState)
        end
        if outputState ~= nil then
            processStateFX(outputState)
        end
        if isDebugEnabled then
            THPlaceable.displayMsg(self, "Active FX objects:")
            THUtils.printTable(fxHandler.activeFXObjects)
        end
        specTable.isFXHandlerDirty = true
        self:raiseActive()
    end
end
function THPlaceableStateControl:thAA_getControlStateStorages(stateTypeIndex)
    local specTable = getSpecTable(self)
    local storageList = nil
    if stateTypeIndex == nil then
        storageList = specTable.activeStorages
    elseif stateTypeIndex == THPlaceableStateControl.STATE_TYPE.INPUT then
        if specTable.unloadingStation ~= nil then
            storageList = specTable.unloadingStation.targetStorages
        end
    elseif stateTypeIndex == THPlaceableStateControl.STATE_TYPE.OUTPUT then
        if specTable.loadingStation ~= nil then
            storageList = specTable.loadingStation.sourceStorages
        end
    else
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "stateTypeIndex", stateTypeIndex)
    end
    return storageList
end
function THPlaceableStateControl:thAA_updateStateControlStorages()
    local specTable = getSpecTable(self)
    local mission = thMain.mission
    THUtils.clearTable(specTable.activeStorages)
    THUtils.clearTable(specTable.activeStoragesArray)
    if mission ~= nil and mission.storageSystem ~= nil
        and self:thAA_getIsStateControlEnabled()
    then
        local ownerFarmId = self:getOwnerFarmId()
        local storageSystem = mission.storageSystem
        local unloadingStation = specTable.unloadingStation
        local loadingStation = specTable.loadingStation
        local isDebugEnabled = THDebugUtil.getIsDebugEnabled(debugFlagId)
        local function updateStateTypeStorage(pOtherStateType)
            local controlState = self:thAA_getCurrentControlState(pOtherStateType)
            local linkedStorageList = specTable.defaultStorages[pOtherStateType]
            local currentStorageList = nil
            if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                currentStorageList = unloadingStation.targetStorages
            elseif pOtherStateType == THPlaceableStateControl.STATE_TYPE.OUTPUT then
                currentStorageList = loadingStation.sourceStorages
            else
                THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "stateTypeIndex", pOtherStateType)
                return
            end
            if controlState ~= nil then
                linkedStorageList = controlState.linkedStorages
            end
            if currentStorageList ~= nil and linkedStorageList ~= nil then
                for storage in pairs(currentStorageList) do
                    if not linkedStorageList[storage] then
                        if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                            storageSystem:removeStorageFromUnloadingStations(storage, { unloadingStation })
                        else
                            storageSystem:removeStorageFromLoadingStations(storage, { loadingStation })
                        end
                    end
                end
                for storage in pairs(linkedStorageList) do
                    if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                        storageSystem:addStorageToUnloadingStation(storage, unloadingStation)
                    else
                        storageSystem:addStorageToLoadingStation(storage, loadingStation)
                    end
                    if not specTable.activeStorages[storage] then
                        table.insert(specTable.activeStoragesArray, storage)
                        specTable.activeStorages[storage] = true
                    end
                end
                if controlState ~= nil then
                    local extStorageArray = nil
                    if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                        unloadingStation.storageRadius = controlState.extensionRange or 0
                        extStorageArray = storageSystem:getStorageExtensionsInRange(unloadingStation, ownerFarmId)
                    else
                        loadingStation.storageRadius = controlState.extensionRange or 0
                        extStorageArray = storageSystem:getStorageExtensionsInRange(loadingStation, ownerFarmId)
                    end
                    if extStorageArray ~= nil then
                        for _, storage in pairs(extStorageArray) do
                            if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                                if not unloadingStation.targetStorages[storage] then
                                    storageSystem:addStorageToUnloadingStation(storage, unloadingStation)
                                end
                            else
                                if not loadingStation.sourceStorages[storage] then
                                    storageSystem:addStorageToLoadingStation(storage, loadingStation)
                                end
                            end
                            if not specTable.activeStorages[storage] then
                                table.insert(specTable.activeStoragesArray, storage)
                                specTable.activeStorages[storage] = true
                            end
                        end
                    end
                end
            end
            if specTable.storageTransfer ~= nil then
                if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                    specTable.storageTransfer:setInputStorages(linkedStorageList)
                else
                    specTable.storageTransfer:setOutputStorages(linkedStorageList)
                end
            end
        end
        if unloadingStation ~= nil and unloadingStation.targetStorages ~= nil then
            updateStateTypeStorage(THPlaceableStateControl.STATE_TYPE.INPUT)
            if isDebugEnabled then
                THPlaceable.displayMsg(self, "Unloading station current storages:")
                THUtils.printTable(unloadingStation.targetStorages)
            end
        end
        if loadingStation ~= nil and loadingStation.sourceStorages ~= nil then
            updateStateTypeStorage(THPlaceableStateControl.STATE_TYPE.OUTPUT)
            if isDebugEnabled then
                THPlaceable.displayMsg(self, "Loading station current storages:")
                THUtils.printTable(loadingStation.sourceStorages)
            end
        end
    end
end
function THPlaceableStateControl:thAA_updateControlStateStorageBuffer(stateTypeIndex)
    local specTable = getSpecTable(self)
    if THUtils.argIsValid(stateTypeIndex == nil or THPlaceableStateControl.STATE_TYPE_ID[stateTypeIndex] ~= nil, "stateTypeIndex", stateTypeIndex) then
        local thMoistureSpec, storageDryer, isDryerRunning = nil, nil, false
        if THPlaceableMoisture ~= nil then
            thMoistureSpec = THPlaceableMoisture.getSpecTable(self)
            if thMoistureSpec ~= nil then
                storageDryer = thMoistureSpec.storageDryer
                if storageDryer ~= nil then
                    isDryerRunning = storageDryer:getCurrentStatus() == THStorageDryer.STATUS.RUNNING
                end
            end
        end
        local bufferMapping = specTable.storageBufferMapping
        local function updateStateTypeBuffer(pOtherStateType)
            local storageBuffer = THUtils.clearTable(specTable.storageBuffer[pOtherStateType])
            local bufferStorages = nil
            THUtils.clearTable(bufferMapping)
            if self:thAA_getIsStateControlEnabled() then
                if pOtherStateType == THPlaceableStateControl.STATE_TYPE.INPUT then
                    if specTable.unloadingStation ~= nil then
                        bufferStorages = specTable.unloadingStation.targetStorages
                    end
                elseif pOtherStateType == THPlaceableStateControl.STATE_TYPE.OUTPUT then
                    if specTable.loadingStation ~= nil then
                        bufferStorages = specTable.loadingStation.sourceStorages
                    end
                end
            end
            if bufferStorages ~= nil then
                for storage in pairs(bufferStorages) do
                    if THUtils.getIsType(storage, Storage) then
                        local storageFillLevels = storage:getFillLevels()
                        if storageFillLevels ~= nil then
                            for storageFillType, storageFillLevel in pairs(storageFillLevels) do
                                local fillTypeInfo = thFillTypeManager:getFillType(storageFillType)
                                if storageFillLevel > 0
                                    and fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN
                                then
                                    local bufferData = storageBuffer[bufferMapping[fillTypeInfo.index]]
                                    if bufferData == nil then
                                        bufferData = {
                                            fillTypeIndex = fillTypeInfo.index,
                                            fillTypeName = fillTypeInfo.name,
                                            fillTypeTitle = fillTypeInfo.title,
                                            fillLevel = 0,
                                            isDrying = false
                                        }
                                        local bufferIndex = #storageBuffer + 1
                                        storageBuffer[bufferIndex] = bufferData
                                        bufferMapping[fillTypeInfo.index] = bufferIndex
                                    end
                                    bufferData.fillLevel = bufferData.fillLevel + storageFillLevel
                                    if isDryerRunning and storageDryer ~= nil
                                        and storageDryer:getIsOutputStorage(storage)
                                    then
                                        bufferData.isDrying = true
                                    end
                                end
                            end
                        end
                    end
                end
            end
            if #storageBuffer > 0 then
                table.sort(storageBuffer, function(pa, pb)
                    return pa.fillLevel < pb.fillLevel
                end)
            end
        end
        if stateTypeIndex == nil then
            for _, otherStateType in pairs(THPlaceableStateControl.STATE_TYPE) do
                updateStateTypeBuffer(otherStateType)
            end
        else
            updateStateTypeBuffer(stateTypeIndex)
        end
    end
end
function THPlaceableStateControl:setOwnerFarmId(superFunc, farmId, noEventSend, ...)
    local specTable = getSpecTable(self)
    local function appendFunc(...)
        THUtils.call(function()
            if specTable.storageTransfer ~= nil then
                specTable.storageTransfer:setOwnerFarmId(farmId, noEventSend)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, farmId, noEventSend, ...))
end
function THPlaceableStateControl:collectPickObjects(superFunc, nodeId, ...)
    local specTable = getSpecTable(self)
    local isNodeFound = false
    if nodeId ~= nil then
        THUtils.call(function()
            if nodeId == specTable.controlTrigger then
                isNodeFound = true
                return
            end
            if specTable.storageTransfer ~= nil then
                if specTable.storageTransfer:getIsPickObjectNode(nodeId) then
                    isNodeFound = true
                    return
                end
            end
        end)
    end
    if not isNodeFound then
        return superFunc(self, nodeId, ...)
    end
end
function THPlaceableStateControl:thAA_updateActionEvents(superFunc, inputContext, ...)
    local function appendFunc(...)
        THUtils.call(function()
            self:thAA_updateStateControlActionEvents(inputContext)
        end)
        return ...
    end
    return appendFunc(superFunc(self, ...))
end
function THPlaceableStateControl:thAA_updateExtendedStorageBuffer(superFunc, ...)
    local rootSpec = THPlaceable.getSpecTable(self)
    local function appendFunc(...)
        THUtils.call(function()
            if self:thAA_getIsStateControlEnabled() then
                THUtils.clearTable(rootSpec.storageBuffer)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, ...))
end
function THPlaceableStateControl:thAA_updateInfoHudEntries(superFunc, infoTable, ...)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        specTable.canActivateTransfer = false
        if self:thAA_getIsStateControlEnabled()
            and self:thAA_getIsActiveForInput()
        then
            local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
            local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
            local function addStateTypeStorageInfo(pOtherStateType)
                local storageBuffer = specTable.storageBuffer[pOtherStateType]
                local isStorageEmpty = true
                if storageBuffer ~= nil then
                    for bufferIndex = 1, #storageBuffer do
                        local bufferData = storageBuffer[bufferIndex]
                        local fillLevel, precision, volumeUnitText = thFillTypeManager:convertFillLevel(bufferData.fillLevel, bufferData.fillTypeIndex)
                        local roundedFillLevel = THUtils.round(fillLevel, precision)
                        if roundedFillLevel > 0 then
                            local fillLevelText = thI18n:formatVolume(roundedFillLevel, precision, volumeUnitText)
                            local fillLevelColor = nil
                            if bufferData.isDrying then
                                fillLevelColor = THHud.COLOR.GREEN
                            end
                            self:thAA_addInfoHudEntry(infoTable, bufferData.fillTypeTitle, fillLevelText, false, fillLevelColor, nil, false, nil, false, fillLevelColor)
                            isStorageEmpty = false
                        end
                    end
                end
                if isStorageEmpty then
                    self:thAA_addInfoHudEntry(infoTable, "", THPlaceableStateControl.HUD_TEXT.STORAGE_EMPTY)
                end
            end
            if inputState ~= nil then
                self:thAA_addInfoHudEntry(infoTable, THPlaceableStateControl.HUD_TEXT.INPUT_STATE, inputState.title, true, THHud.COLOR.BLUE, 1.2, true, nil, true)
                addStateTypeStorageInfo(THPlaceableStateControl.STATE_TYPE.INPUT)
            end
            if outputState ~= nil then
                self:thAA_addInfoHudEntry(infoTable, THPlaceableStateControl.HUD_TEXT.OUTPUT_STATE, outputState.title, true, THHud.COLOR.BLUE, 1.2, true, nil, true)
                addStateTypeStorageInfo(THPlaceableStateControl.STATE_TYPE.OUTPUT)
            end
            if specTable.storageTransfer ~= nil then
                local transferStatus = specTable.storageTransfer:getCurrentStatus()
                if transferStatus ~= THStorageTransfer.STATUS.DISABLED
                    and inputState ~= nil and outputState ~= nil
                then
                    local transferStatusText = specTable.storageTransfer:getStatusText(transferStatus)
                    local transferStatusColor = nil
                    if transferStatus == THStorageTransfer.STATUS.RUNNING then
                        transferStatusColor = THHud.COLOR.GREEN
                    elseif transferStatus == THStorageTransfer.STATUS.STARTING
                        or transferStatus == THStorageTransfer.STATUS.STOPPING
                    then
                        transferStatusColor = THHud.COLOR.ORANGE
                    end
                    specTable.canActivateTransfer = true
                    self:thAA_addInfoHudEntry(infoTable, THPlaceableStateControl.HUD_TEXT.TRANSFER_STATUS, transferStatusText, true, THHud.COLOR.ORANGE, nil, false, nil, false, transferStatusColor)
                end
            end
        end
    end)
    return superFunc(self, infoTable, ...)
end
function THPlaceableStateControl:thAA_updateMoistureStorageBuffer(superFunc, storageList, storageBuffer, ...)
    local specTable = getSpecTable(self)
    local function appendFunc(rOtherBuffer, ...)
        THUtils.call(function()
            if self:thAA_getIsStateControlEnabled() then
                local thMoistureSpec = THPlaceableMoisture.getSpecTable(self)
                if rOtherBuffer == thMoistureSpec.storageBuffer then
                    for _, stateTypeIndex in pairs(THPlaceableStateControl.STATE_TYPE) do
                        local otherStorageList = self:thAA_getControlStateStorages(stateTypeIndex)
                        if otherStorageList ~= nil then
                            superFunc(self, otherStorageList, specTable.moistureStorageBuffer[stateTypeIndex])
                        end
                    end
                    local storageDryer = thMoistureSpec.storageDryer
                    if storageDryer ~= nil then
                        local dryerStorageList = storageDryer:getOutputStorages()
                        if dryerStorageList ~= nil then
                            superFunc(self, dryerStorageList, specTable.dryerMoistureBuffer)
                        end
                    end
                end
            end
        end)
        return rOtherBuffer, ...
    end
    return appendFunc(superFunc(self, storageList, storageBuffer, ...))
end
function THPlaceableStateControl:thAA_addMoistureStorageInfoHudEntries(superFunc, infoTable, hudTitle, storageBuffer, ...)
    local specTable = getSpecTable(self)
    local hasInputBuffer, isDefaultBuffer = false, false
    THUtils.call(function()
        if self:thAA_getIsStateControlEnabled()
            and self:thAA_getIsActiveForInput()
        then
            local thMoistureSpec = THPlaceableMoisture.getSpecTable(self)
            if storageBuffer == nil or storageBuffer == thMoistureSpec.storageBuffer then
                local inputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.INPUT)
                local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
                local inputStorageBuffer = specTable.moistureStorageBuffer[THPlaceableStateControl.STATE_TYPE.INPUT]
                local outputStorageBuffer = specTable.moistureStorageBuffer[THPlaceableStateControl.STATE_TYPE.OUTPUT]
                if inputState ~= nil and #inputStorageBuffer > 0
                    and (outputState == nil or inputState ~= outputState)
                then
                    hudTitle = THPlaceableStateControl.HUD_TEXT.INPUT_MOISTURE
                    storageBuffer = inputStorageBuffer
                    hasInputBuffer = true
                elseif outputState ~= nil and #outputStorageBuffer > 0 then
                    hudTitle = THPlaceableStateControl.HUD_TEXT.OUTPUT_MOISTURE
                    storageBuffer = outputStorageBuffer
                end
                isDefaultBuffer = true
            end
        end
    end)
    local function appendFunc(rSuccess, rHasMoisture, rHasLoss, rCanDry, ...)
        if rSuccess and isDefaultBuffer then
            THUtils.call(function()
                if hasInputBuffer then
                    local outputState = self:thAA_getCurrentControlState(THPlaceableStateControl.STATE_TYPE.OUTPUT)
                    local outputStorageBuffer = specTable.moistureStorageBuffer[THPlaceableStateControl.STATE_TYPE.OUTPUT]
                    if outputState ~= nil and #outputStorageBuffer > 0 then
                        local _, otherHasMoisture, otherHasLoss = superFunc(self, infoTable, THPlaceableStateControl.HUD_TEXT.OUTPUT_MOISTURE, outputStorageBuffer)
                        rHasMoisture = rHasMoisture or otherHasMoisture == true
                        rHasLoss = rHasLoss or otherHasLoss == true
                    end
                end
                rCanDry = false
                if #specTable.dryerMoistureBuffer > 0 then
                    local _, otherHasMoisture, otherHasLoss, otherCanDry = superFunc(self, infoTable, THPlaceableStateControl.HUD_TEXT.DRYER_MOISTURE, specTable.dryerMoistureBuffer)
                    rHasMoisture = rHasMoisture or otherHasMoisture == true
                    rHasLoss = rHasLoss or otherHasLoss == true
                    rCanDry = otherCanDry == true
                end
            end)
        end
        return rSuccess, rHasMoisture, rHasLoss, rCanDry, ...
    end
    return appendFunc(superFunc(self, infoTable, hudTitle, storageBuffer, ...))
end
function THPlaceableStateControl.inj_loadProductionPoint(superFunc, extraArgs, productionPoint, components, xmlFile, xmlKey, customEnv, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self)
    if specTable ~= nil and xmlFile ~= nil then
        THUtils.call(function()
            local specXMLKey = THPlaceable.getConfigXMLKey(THPlaceableStateControl)
            local productionPointKey = string.format("%s.productionPoint", specXMLKey)
            if xmlFile:hasProperty(specXMLKey) then
                xmlKey = productionPointKey
            end
        end)
    end
    return superFunc(productionPoint, components, xmlFile, xmlKey, customEnv, ...)
end
function THSelectControlStateEvent.emptyNew()
    return Event.new(THSelectControlStateEvent_mt)
end
function THSelectControlStateEvent.new(target, stateTypeIndex)
    local self = THSelectControlStateEvent.emptyNew()
    self.isEventValid = false
    if THUtils.argIsValid(THUtils.getIsType(target, Placeable), "target", target)
        and THUtils.argIsValid(THPlaceableStateControl.STATE_TYPE_ID[stateTypeIndex] ~= nil, "stateTypeIndex", stateTypeIndex)
    then
        self.target = target
        self.stateTypeIndex = stateTypeIndex
        self.isEventValid = true
    end
    return self
end
function THSelectControlStateEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        self.isEventValid = streamReadBool(streamId)
        if self.isEventValid then
            self.target = NetworkUtil.readNodeObject(streamId)
            self.stateTypeIndex = streamReadUIntN(streamId, THPlaceableStateControl.STATE_TYPE_BITS)
            self:run(connection)
        end
    else
        THUtils.errorMsg(true, "Event should not read from clients!")
    end
end
function THSelectControlStateEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        if streamWriteBool(streamId, self.isEventValid) then
            NetworkUtil.writeNodeObject(streamId, self.target)
            streamWriteUIntN(streamId, self.stateTypeIndex, THPlaceableStateControl.STATE_TYPE_BITS)
        end
    else
        THUtils.errorMsg(true, "Event should not write from server!")
    end
end
function THSelectControlStateEvent:run(connection)
    if not connection:getIsServer() then -- server reading from client
        if self.isEventValid then
            local specTable = getSpecTable(self.target)
            if specTable ~= nil then
                self.target:thAA_selectControlState(self.stateTypeIndex)
            end
        end
    end
end